"""Index entries adapters for sphinx.environment."""

from __future__ import annotations

import re
import unicodedata
from itertools import groupby
from typing import TYPE_CHECKING

from sphinx.errors import NoUri
from sphinx.locale import _, __
from sphinx.util import logging
from sphinx.util.index_entries import _split_into

if TYPE_CHECKING:
    from typing import Literal

    from sphinx.builders import Builder
    from sphinx.environment import BuildEnvironment

    type _IndexEntryTarget = tuple[str | None, str | Literal[False]]
    type _IndexEntryTargets = list[_IndexEntryTarget]
    type _IndexEntryCategoryKey = str | None
    type _IndexEntrySubItems = dict[
        str,
        tuple[_IndexEntryTargets, _IndexEntryCategoryKey],
    ]
    type _IndexEntry = tuple[
        _IndexEntryTargets,
        _IndexEntrySubItems,
        _IndexEntryCategoryKey,
    ]
    type _IndexEntryMap = dict[str, _IndexEntry]

    # Used by ``create_index()`` for 'the real index'
    type _RealIndexEntry = tuple[
        str,
        tuple[
            _IndexEntryTargets,
            list[tuple[str, _IndexEntryTargets]],
            _IndexEntryCategoryKey,
        ],
    ]
    type _RealIndexEntries = list[_RealIndexEntry]
    type _Index = list[tuple[str, _RealIndexEntries]]

logger = logging.getLogger(__name__)


class IndexEntries:
    def __init__(self, env: BuildEnvironment) -> None:
        self.env = env

    def create_index(
        self,
        builder: Builder,
        group_entries: bool = True,
        _fixre: re.Pattern[str] = re.compile(r'(.*) ([(][^()]*[)])'),
    ) -> _Index:
        """Create the real index from the collected index entries."""
        new: _IndexEntryMap = {}

        rel_uri: str | Literal[False]
        index_domain = self.env.domains.index_domain
        for docname, entries in index_domain.entries.items():
            try:
                rel_uri = builder.get_relative_uri('genindex', docname)
            except NoUri:
                rel_uri = False

            # new entry types must be listed in directives/other.py!
            for entry_type, value, target_id, main, category_key in entries:
                uri = rel_uri is not False and f'{rel_uri}#{target_id}'
                try:
                    if entry_type == 'single':
                        try:
                            entry, sub_entry = _split_into(2, 'single', value)
                        except ValueError:
                            (entry,) = _split_into(1, 'single', value)
                            sub_entry = ''
                        _add_entry(
                            entry, sub_entry, main, dic=new, link=uri, key=category_key
                        )
                    elif entry_type == 'pair':
                        first, second = _split_into(2, 'pair', value)
                        _add_entry(
                            first, second, main, dic=new, link=uri, key=category_key
                        )
                        _add_entry(
                            second, first, main, dic=new, link=uri, key=category_key
                        )
                    elif entry_type == 'triple':
                        first, second, third = _split_into(3, 'triple', value)
                        _add_entry(
                            first,
                            second + ' ' + third,
                            main,
                            dic=new,
                            link=uri,
                            key=category_key,
                        )
                        _add_entry(
                            second,
                            third + ', ' + first,
                            main,
                            dic=new,
                            link=uri,
                            key=category_key,
                        )
                        _add_entry(
                            third,
                            first + ' ' + second,
                            main,
                            dic=new,
                            link=uri,
                            key=category_key,
                        )
                    elif entry_type == 'see':
                        first, second = _split_into(2, 'see', value)
                        _add_entry(
                            first,
                            _('see %s') % second,
                            None,
                            dic=new,
                            link=False,
                            key=category_key,
                        )
                    elif entry_type == 'seealso':
                        first, second = _split_into(2, 'see', value)
                        _add_entry(
                            first,
                            _('see also %s') % second,
                            None,
                            dic=new,
                            link=False,
                            key=category_key,
                        )
                    else:
                        logger.warning(
                            __('unknown index entry type %r'),
                            entry_type,
                            location=docname,
                            type='index',
                        )
                except ValueError as err:
                    logger.warning(str(err), location=docname, type='index')

        for targets, sub_items, _category_key in new.values():
            targets.sort(key=_key_func_0)
            for sub_targets, _sub_category_key in sub_items.values():
                sub_targets.sort(key=_key_func_0)

        new_list: list[tuple[str, _IndexEntry]] = sorted(new.items(), key=_key_func_1)

        if group_entries:
            # fixup entries: transform
            #   func() (in module foo)
            #   func() (in module bar)
            # into
            #   func()
            #     (in module foo)
            #     (in module bar)
            old_key = ''
            old_sub_items: _IndexEntrySubItems = {}
            i = 0
            while i < len(new_list):
                key, (targets, sub_items, category_key) = new_list[i]
                # cannot move if it has sub_items; structure gets too complex
                if not sub_items:
                    m = _fixre.match(key)
                    if m:
                        if old_key == m.group(1):
                            # prefixes match: add entry as subitem of the
                            # previous entry
                            prev = old_sub_items.setdefault(m[2], ([], category_key))
                            prev[0].extend(targets)
                            del new_list[i]
                            continue
                        old_key = m.group(1)
                    else:
                        old_key = key
                old_sub_items = sub_items
                i += 1

        grouped = []
        for group_key, group in groupby(new_list, _group_by_func):
            group_list = []
            for group_entry in group:
                entry_key, (targets, sub_items, category_key) = group_entry
                pairs = [
                    (sub_key, sub_targets)
                    for (sub_key, (sub_targets, _sub_category_key)) in sub_items.items()
                ]
                pairs.sort(key=_key_func_2)
                group_list.append((entry_key, (targets, pairs, category_key)))
            grouped.append((group_key, group_list))
        return grouped


def _add_entry(
    word: str,
    subword: str,
    main: str | None,
    *,
    dic: _IndexEntryMap,
    link: str | Literal[False],
    key: _IndexEntryCategoryKey,
) -> None:
    entry = dic.setdefault(word, ([], {}, key))
    if subword:
        targets = entry[1].setdefault(subword, ([], key))[0]
    else:
        targets = entry[0]
    if link:
        targets.append((main, link))


def _key_func_0(entry: _IndexEntryTarget) -> tuple[bool, str | Literal[False]]:
    """Sort the index entries for same keyword."""
    main, uri = entry
    return not main, uri  # show main entries at first


def _key_func_1(entry: tuple[str, _IndexEntry]) -> tuple[tuple[int, str], str]:
    """Sort the index entries"""
    key, (_targets, _sub_items, category_key) = entry
    if category_key:
        # using the specified category key to sort
        key = category_key
    lc_key = unicodedata.normalize('NFD', key.lower())
    lc_key = lc_key.removeprefix('\N{RIGHT-TO-LEFT MARK}')

    if not lc_key[0:1].isalpha() and not lc_key.startswith('_'):
        # put symbols at the front of the index (0)
        group = 0
    else:
        # put non-symbol characters at the following group (1)
        group = 1
    # ensure a deterministic order *within* letters by also sorting on
    # the entry itself
    return (group, lc_key), entry[0]


def _key_func_2(entry: tuple[str, _IndexEntryTargets]) -> str:
    """Sort the sub-index entries"""
    key = unicodedata.normalize('NFD', entry[0].lower())
    key = key.removeprefix('\N{RIGHT-TO-LEFT MARK}')
    if key[0:1].isalpha() or key.startswith('_'):
        key = chr(127) + key
    return key


def _group_by_func(entry: tuple[str, _IndexEntry]) -> str:
    """Group the entries by letter or category key."""
    key, (_targets, _sub_items, category_key) = entry

    if category_key is not None:
        return category_key

    # now calculate the key
    key = key.removeprefix('\N{RIGHT-TO-LEFT MARK}')
    letter = unicodedata.normalize('NFD', key[0])[0].upper()
    if letter.isalpha() or letter == '_':
        return letter

    # get all other symbols under one heading
    return _('Symbols')



from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx.ext.autodoc._property_types import (
    _AssignStatementProperties,
    _ClassDefProperties,
    _FunctionDefProperties,
    _TypeStatementProperties,
)
from sphinx.ext.autodoc._sentinels import SUPPRESS
from sphinx.locale import _

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Literal

    from docutils.statemachine import StringList

    from sphinx.ext.autodoc._directive_options import _AutoDocumenterOptions
    from sphinx.ext.autodoc._property_types import _ItemProperties


def _directive_header_lines(
    *,
    autodoc_typehints: Literal['signature', 'description', 'none', 'both'],
    directive_name: str,
    is_final: bool,
    options: _AutoDocumenterOptions,
    props: _ItemProperties,
) -> Iterator[str]:
    """Emit the directive header and option lines."""
    # normally the name doesn't contain the module
    # (except for module directives of course)
    name = props.dotted_parts or props.module_name

    # emit one signature per line
    # the first line contains the directive prefix
    sig_line, *sig_lines = props.signatures or ('',)
    prefix = f'.. {directive_name}:: '
    yield f'{prefix}{name}{sig_line}'
    # emit remaining lines, indented to the same column
    prefix = ' ' * len(prefix)
    for sig_line in sig_lines:
        yield f'{prefix}{name}{sig_line}'

    if options.no_index or options.noindex:
        yield '   :no-index:'
    if options.no_index_entry:
        yield '   :no-index-entry:'
    if props.parts:
        # Be explicit about the module, this is necessary since .. class::
        # etc. don't support a prepended module name
        yield f'   :module: {props.module_name}'

    if props.obj_type == 'module':
        # add some module-specific options
        if options.synopsis:
            yield f'   :synopsis: {options.synopsis}'
        if options.platform:
            yield f'   :platform: {options.platform}'
        if options.deprecated:
            yield '   :deprecated:'

    if props.obj_type in {'class', 'exception'}:
        assert isinstance(props, _ClassDefProperties)

        if props._obj_is_new_type or props._obj_is_typevar:
            return

        if is_final:
            yield '   :final:'

        canonical_fullname = props.canonical_full_name
        if (
            not props.doc_as_attr
            and not props._obj_is_new_type
            and canonical_fullname
            and props.full_name != canonical_fullname
        ):
            yield f'   :canonical: {canonical_fullname}'

        # add inheritance info, if wanted
        if not props.doc_as_attr and options.show_inheritance:
            yield ''
            yield '   ' + _('Bases: %s') % ', '.join(props._obj_bases)

    if props.obj_type in {'function', 'decorator'}:
        assert isinstance(props, _FunctionDefProperties)

        if props.is_async:
            yield '   :async:'

    if props.obj_type == 'method':
        assert isinstance(props, _FunctionDefProperties)

        if props.is_abstractmethod:
            yield '   :abstractmethod:'
        if props.is_async:
            yield '   :async:'
        if props.is_classmethod:
            yield '   :classmethod:'
        if props.is_staticmethod:
            yield '   :staticmethod:'
        if props.is_final or is_final:
            yield '   :final:'

    if props.obj_type == 'property':
        assert isinstance(props, _FunctionDefProperties)

        if props.is_abstractmethod:
            yield '   :abstractmethod:'
        if props.is_classmethod:
            yield '   :classmethod:'

        objrepr = props._obj_property_type_annotation
        if autodoc_typehints != 'none' and objrepr is not None:
            yield f'   :type: {objrepr}'

    if props.obj_type == 'data':
        assert isinstance(props, _AssignStatementProperties)

        if options.annotation is SUPPRESS or props._obj_is_generic_alias:
            pass
        elif options.annotation:
            yield f'   :annotation: {options.annotation}'
        else:
            type_annotation = props._obj_type_annotation
            if autodoc_typehints != 'none' and type_annotation is not None:
                yield f'   :type: {type_annotation}'

            if (
                not options.no_value
                and props._obj_is_sentinel is None  # not any sentinel
                and not props._docstrings_has_hide_value
                and not props._obj_is_mock
            ):
                yield f'   :value: {props._obj_repr_rst}'

    if props.obj_type == 'attribute':
        assert isinstance(props, _AssignStatementProperties)

        if (
            options.annotation
            and options.annotation is not SUPPRESS
            and not props._obj_is_generic_alias
        ):
            yield f'   :annotation: {options.annotation}'
        else:
            type_annotation = props._obj_type_annotation
            if autodoc_typehints != 'none' and type_annotation is not None:
                yield f'   :type: {type_annotation}'

            if (
                not options.no_value
                and props._obj_is_sentinel is None  # not any sentinel
                and not props._obj_is_attribute_descriptor
                and not props._obj_is_generic_alias
                and not props._docstrings_has_hide_value
                and not props._obj_is_mock
            ):
                yield f'   :value: {props._obj_repr_rst}'

    if props.obj_type == 'type':
        assert isinstance(props, _TypeStatementProperties)

        if not options.no_value and not props._docstrings_has_hide_value:
            yield f'   :canonical: {props._obj___value__}'


def _add_content(content: StringList, *, result: StringList, indent: str) -> None:
    for line, src in zip(content.data, content.items, strict=True):
        if line.strip():  # not a blank line
            result.append(indent + line, src[0], src[1])
        else:
            result.append('', src[0], src[1])



"""Allow todos to be inserted into your documentation.

Inclusion of todos can be switched of by a configuration variable.
The todolist directive collects all todos of your project and lists them along
with a backlink to the original location.
"""

from __future__ import annotations

import functools
import operator
from typing import TYPE_CHECKING, cast

from docutils import nodes

import sphinx
from sphinx import addnodes
from sphinx.directives.admonitions import SphinxAdmonition
from sphinx.domains import Domain
from sphinx.errors import NoUri
from sphinx.locale import _, __
from sphinx.util import logging, texescape
from sphinx.util.docutils import SphinxDirective, new_document

if TYPE_CHECKING:
    from collections.abc import Set
    from typing import Any, ClassVar

    from docutils.nodes import Element, Node

    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment
    from sphinx.util.typing import ExtensionMetadata, OptionSpec
    from sphinx.writers.html5 import HTML5Translator
    from sphinx.writers.latex import LaTeXTranslator

logger = logging.getLogger(__name__)


class todo_node(nodes.Admonition, nodes.Element):
    pass


class todolist(nodes.General, nodes.Element):
    pass


class Todo(SphinxAdmonition):
    """A todo entry, displayed (if configured) in the form of an admonition."""

    node_class = todo_node

    def run(self) -> list[Node]:
        if not self.options.get('class'):
            self.options['class'] = ['admonition-todo']

        (todo,) = super().run()
        if not isinstance(todo, todo_node):
            return [todo]

        todo.insert(0, nodes.title(text=_('Todo')))
        todo['docname'] = self.env.current_document.docname
        self.add_name(todo)
        self.set_source_info(todo)
        self.state.document.note_explicit_target(todo)
        return [todo]


class TodoDomain(Domain):
    name = 'todo'
    label = 'todo'

    @property
    def todos(self) -> dict[str, list[todo_node]]:
        return self.data.setdefault('todos', {})

    def clear_doc(self, docname: str) -> None:
        self.todos.pop(docname, None)

    def merge_domaindata(self, docnames: Set[str], otherdata: dict[str, Any]) -> None:
        for docname in docnames:
            self.todos[docname] = otherdata['todos'][docname]

    def process_doc(
        self, env: BuildEnvironment, docname: str, document: nodes.document
    ) -> None:
        todos = self.todos.setdefault(docname, [])
        for todo in document.findall(todo_node):
            env.events.emit('todo-defined', todo)
            todos.append(todo)

            if env.config.todo_emit_warnings:
                logger.warning(
                    __('TODO entry found: %s'), todo[1].astext(), location=todo
                )


class TodoList(SphinxDirective):
    """A list of all todo entries."""

    has_content = False
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec: ClassVar[OptionSpec] = {}

    def run(self) -> list[Node]:
        # Simply insert an empty todolist node which will be replaced later
        # when process_todo_nodes is called
        return [todolist('')]


class TodoListProcessor:
    def __init__(self, app: Sphinx, doctree: nodes.document, docname: str) -> None:
        self.builder = app.builder
        self.config = app.config
        self.env = app.env
        self.domain = app.env.domains['todo']
        self.document = new_document('')

        self.process(doctree, docname)

    def process(self, doctree: nodes.document, docname: str) -> None:
        todos: list[todo_node] = functools.reduce(
            operator.iadd, self.domain.todos.values(), []
        )
        for node in list(doctree.findall(todolist)):
            if not self.config.todo_include_todos:
                node.parent.remove(node)
                continue

            if node.get('ids'):
                content: list[Element] = [nodes.target()]
            else:
                content = []

            for todo in todos:
                # Create a copy of the todo node
                new_todo = todo.deepcopy()
                new_todo['ids'].clear()

                self.resolve_reference(new_todo, docname)
                content.append(new_todo)

                todo_ref = self.create_todo_reference(todo, docname)
                content.append(todo_ref)

            node.replace_self(content)

    def create_todo_reference(self, todo: todo_node, docname: str) -> nodes.paragraph:
        if self.config.todo_link_only:
            description = _('<<original entry>>')
        else:
            description = _('(The <<original entry>> is located in %s, line %d.)') % (
                todo.source,
                todo.line,
            )

        prefix = description[: description.find('<<')]
        suffix = description[description.find('>>') + 2 :]

        para = nodes.paragraph(classes=['todo-source'])
        para += nodes.Text(prefix)

        # Create a reference
        linktext = nodes.emphasis(_('original entry'), _('original entry'))
        reference = nodes.reference('', '', linktext, internal=True)
        try:
            reference['refuri'] = self.builder.get_relative_uri(
                docname, todo['docname']
            )
            reference['refuri'] += '#' + todo['ids'][0]
        except NoUri:
            # ignore if no URI can be determined, e.g. for LaTeX output
            pass

        para += reference
        para += nodes.Text(suffix)

        return para

    def resolve_reference(self, todo: todo_node, docname: str) -> None:
        """Resolve references in the todo content."""
        for node in todo.findall(addnodes.pending_xref):
            if 'refdoc' in node:
                node['refdoc'] = docname

        # Note: To resolve references, it is needed to wrap it with document node
        self.document += todo
        self.env.resolve_references(self.document, docname, self.builder)
        self.document.remove(todo)


def visit_todo_node(self: HTML5Translator, node: todo_node) -> None:
    if self.config.todo_include_todos:
        self.visit_admonition(node)
    else:
        raise nodes.SkipNode


def depart_todo_node(self: HTML5Translator, node: todo_node) -> None:
    self.depart_admonition(node)


def latex_visit_todo_node(self: LaTeXTranslator, node: todo_node) -> None:
    if self.config.todo_include_todos:
        self.body.append('\n\\begin{sphinxtodo}{')
        self.body.append(self.hypertarget_to(node))

        title_node = cast('nodes.title', node[0])
        title = texescape.escape(title_node.astext(), self.config.latex_engine)
        self.body.append('%s:}' % title)
        self.no_latex_floats += 1
        if self.table:
            self.table.has_problematic = True
        node.pop(0)
    else:
        raise nodes.SkipNode


def latex_depart_todo_node(self: LaTeXTranslator, node: todo_node) -> None:
    self.body.append('\\end{sphinxtodo}\n')
    self.no_latex_floats -= 1


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_event('todo-defined')
    app.add_config_value('todo_include_todos', False, 'html', types=frozenset({bool}))
    app.add_config_value('todo_link_only', False, 'html', types=frozenset({bool}))
    app.add_config_value('todo_emit_warnings', False, 'html', types=frozenset({bool}))

    app.add_node(todolist)
    app.add_node(
        todo_node,
        html=(visit_todo_node, depart_todo_node),
        latex=(latex_visit_todo_node, latex_depart_todo_node),
        text=(visit_todo_node, depart_todo_node),
        man=(visit_todo_node, depart_todo_node),
        texinfo=(visit_todo_node, depart_todo_node),
    )

    app.add_directive('todo', Todo)
    app.add_directive('todolist', TodoList)
    app.add_domain(TodoDomain)
    app.connect('doctree-resolved', TodoListProcessor)
    return {
        'version': sphinx.__display_version__,
        'env_version': 2,
        'parallel_read_safe': True,
    }



from __future__ import annotations

import ast
import functools
import operator
import token
from collections import deque
from inspect import Parameter
from itertools import chain, islice
from typing import TYPE_CHECKING

from docutils import nodes

from sphinx import addnodes
from sphinx.addnodes import pending_xref, pending_xref_condition
from sphinx.locale import _
from sphinx.pycode.parser import Token, TokenProcessor
from sphinx.util.inspect import signature_from_str

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from typing import Any

    from docutils.nodes import Element, Node

    from sphinx.addnodes import desc_signature
    from sphinx.environment import BuildEnvironment


def parse_reftarget(
    reftarget: str, suppress_prefix: bool = False
) -> tuple[str, str, str, bool]:
    """Parse a type string and return (reftype, reftarget, title, refspecific flag)"""
    refspecific = False
    if reftarget.startswith('.'):
        reftarget = reftarget[1:]
        title = reftarget
        refspecific = True
    elif reftarget.startswith('~'):
        reftarget = reftarget[1:]
        title = reftarget.split('.')[-1]
    elif suppress_prefix:
        title = reftarget.split('.')[-1]
    elif reftarget.startswith('typing.'):
        title = reftarget[7:]
    else:
        title = reftarget

    if reftarget == 'None' or reftarget.startswith('typing.'):
        # typing module provides non-class types.  Obj reference is good to refer them.
        reftype = 'obj'
    else:
        reftype = 'class'

    return reftype, reftarget, title, refspecific


def type_to_xref(
    target: str, env: BuildEnvironment, *, suppress_prefix: bool = False
) -> addnodes.pending_xref:
    """Convert a type string to a cross reference node."""
    if env:
        kwargs = {
            'py:module': env.ref_context.get('py:module'),
            'py:class': env.ref_context.get('py:class'),
        }
    else:
        kwargs = {}

    reftype, target, title, refspecific = parse_reftarget(target, suppress_prefix)

    if env.config.python_use_unqualified_type_names:
        # Note: It would be better to use qualname to describe the object to support support
        # nested classes.  But python domain can't access the real python object because this
        # module should work not-dynamically.
        shortname = title.split('.')[-1]
        contnodes: list[Node] = [
            pending_xref_condition('', shortname, condition='resolved'),
            pending_xref_condition('', title, condition='*'),
        ]
    else:
        contnodes = [nodes.Text(title)]

    return pending_xref(
        '',
        *contnodes,
        refdomain='py',
        reftype=reftype,
        reftarget=target,
        refspecific=refspecific,
        **kwargs,
    )


def _parse_annotation(annotation: str, env: BuildEnvironment) -> list[Node]:
    """Parse type annotation."""
    short_literals = env.config.python_display_short_literal_types

    def unparse(node: ast.AST) -> list[Node]:
        if isinstance(node, ast.Attribute):
            return [nodes.Text(f'{unparse(node.value)[0]}.{node.attr}')]
        if isinstance(node, ast.BinOp):
            result: list[Node] = unparse(node.left)
            result.extend(unparse(node.op))
            result.extend(unparse(node.right))
            return result
        if isinstance(node, ast.BitOr):
            return [
                addnodes.desc_sig_space(),
                addnodes.desc_sig_punctuation('', '|'),
                addnodes.desc_sig_space(),
            ]
        if isinstance(node, ast.Constant):
            if node.value is Ellipsis:
                return [addnodes.desc_sig_punctuation('', '...')]
            if isinstance(node.value, bool):
                return [addnodes.desc_sig_keyword('', repr(node.value))]
            if isinstance(node.value, int):
                return [addnodes.desc_sig_literal_number('', repr(node.value))]
            if isinstance(node.value, str):
                return [addnodes.desc_sig_literal_string('', repr(node.value))]
            else:
                # handles None, which is further handled by type_to_xref later
                # and fallback for other types that should be converted
                return [nodes.Text(repr(node.value))]
        if isinstance(node, ast.Expr):
            return unparse(node.value)
        if isinstance(node, ast.Starred):
            result = [addnodes.desc_sig_operator('', '*')]
            result.extend(unparse(node.value))
            return result
        if isinstance(node, ast.Invert):
            return [addnodes.desc_sig_punctuation('', '~')]
        if isinstance(node, ast.USub):
            return [addnodes.desc_sig_punctuation('', '-')]
        if isinstance(node, ast.List):
            result = [addnodes.desc_sig_punctuation('', '[')]
            if node.elts:
                # check if there are elements in node.elts to only pop the
                # last element of result if the for-loop was run at least
                # once
                for elem in node.elts:
                    result.extend(unparse(elem))
                    result.append(addnodes.desc_sig_punctuation('', ','))
                    result.append(addnodes.desc_sig_space())
                result.pop()
                result.pop()
            result.append(addnodes.desc_sig_punctuation('', ']'))
            return result
        if isinstance(node, ast.Module):
            return functools.reduce(operator.iadd, (unparse(e) for e in node.body), [])
        if isinstance(node, ast.Name):
            return [nodes.Text(node.id)]
        if isinstance(node, ast.Subscript):
            if getattr(node.value, 'id', '') in {'Optional', 'Union'}:
                return _unparse_pep_604_annotation(node)
            if short_literals and getattr(node.value, 'id', '') == 'Literal':
                return _unparse_pep_604_annotation(node)
            result = unparse(node.value)
            result.append(addnodes.desc_sig_punctuation('', '['))
            result.extend(unparse(node.slice))
            result.append(addnodes.desc_sig_punctuation('', ']'))

            # Wrap the Text nodes inside brackets by literal node if the subscript is a Literal
            if result[0] in {'Literal', 'typing.Literal'}:
                for i, subnode in enumerate(result[1:], start=1):
                    if isinstance(subnode, nodes.Text):
                        result[i] = nodes.literal('', '', subnode)
            return result
        if isinstance(node, ast.UnaryOp):
            return unparse(node.op) + unparse(node.operand)
        if isinstance(node, ast.Tuple):
            if node.elts:
                result = []
                for elem in node.elts:
                    result.extend(unparse(elem))
                    result.append(addnodes.desc_sig_punctuation('', ','))
                    result.append(addnodes.desc_sig_space())
                result.pop()
                result.pop()
            else:
                result = [
                    addnodes.desc_sig_punctuation('', '('),
                    addnodes.desc_sig_punctuation('', ')'),
                ]

            return result
        if isinstance(node, ast.Call):
            # Call nodes can be used in Annotated type metadata,
            # for example Annotated[str, ArbitraryTypeValidator(str, len=10)]
            args = []
            for arg in node.args:
                args += unparse(arg)
                args.append(addnodes.desc_sig_punctuation('', ','))
                args.append(addnodes.desc_sig_space())
            for kwd in node.keywords:
                args.append(addnodes.desc_sig_name(kwd.arg, kwd.arg))  # type: ignore[arg-type]
                args.append(addnodes.desc_sig_operator('', '='))
                args += unparse(kwd.value)
                args.append(addnodes.desc_sig_punctuation('', ','))
                args.append(addnodes.desc_sig_space())
            result = [
                *unparse(node.func),
                addnodes.desc_sig_punctuation('', '('),
                *args[:-2],  # skip the final comma and space
                addnodes.desc_sig_punctuation('', ')'),
            ]
            return result
        msg = f'unsupported syntax: {node}'
        raise SyntaxError(msg)  # unsupported syntax

    def _unparse_pep_604_annotation(node: ast.Subscript) -> list[Node]:
        subscript = node.slice

        flattened: list[Node] = []
        if isinstance(subscript, ast.Tuple):
            flattened.extend(unparse(subscript.elts[0]))
            for elt in subscript.elts[1:]:
                flattened.extend(unparse(ast.BitOr()))
                flattened.extend(unparse(elt))
        else:
            # e.g. a Union[] inside an Optional[]
            flattened.extend(unparse(subscript))

        if getattr(node.value, 'id', '') == 'Optional':
            flattened.extend(unparse(ast.BitOr()))
            flattened.append(nodes.Text('None'))

        return flattened

    try:
        tree = ast.parse(annotation, type_comments=True)
        result: list[Node] = []
        for node in unparse(tree):
            if isinstance(node, nodes.literal):
                result.append(node[0])
            elif isinstance(node, nodes.Text) and node.strip():
                if (
                    result
                    and isinstance(result[-1], addnodes.desc_sig_punctuation)
                    and result[-1].astext() == '~'
                ):
                    result.pop()
                    result.append(type_to_xref(str(node), env, suppress_prefix=True))
                else:
                    result.append(type_to_xref(str(node), env))
            else:
                result.append(node)
        return result
    except SyntaxError:
        return [type_to_xref(annotation, env)]


class _TypeParameterListParser(TokenProcessor):
    def __init__(self, sig: str) -> None:
        signature = sig.replace('\n', '').strip()
        super().__init__([signature])
        # Each item is a tuple (name, kind, default, annotation) mimicking
        # ``inspect.Parameter`` to allow default values on VAR_POSITIONAL
        # or VAR_KEYWORD parameters.
        self.type_params: list[tuple[str, int, Any, Any]] = []

    def fetch_type_param_spec(self) -> list[Token]:
        tokens = []
        while current := self.fetch_token():
            tokens.append(current)
            for ldelim, rdelim in ('(', ')'), ('{', '}'), ('[', ']'):
                if current == [token.OP, ldelim]:
                    tokens += self.fetch_until([token.OP, rdelim])
                    break
            else:
                if current == token.INDENT:
                    tokens += self.fetch_until(token.DEDENT)
                elif current.match([token.OP, ':'], [token.OP, '='], [token.OP, ',']):
                    tokens.pop()
                    break
        return tokens

    def parse(self) -> None:
        while current := self.fetch_token():
            if current == token.NAME:
                tp_name = current.value.strip()
                if self.previous and self.previous.match(
                    [token.OP, '*'], [token.OP, '**']
                ):
                    if self.previous == [token.OP, '*']:
                        tp_kind = Parameter.VAR_POSITIONAL
                    else:
                        tp_kind = Parameter.VAR_KEYWORD  # type: ignore[assignment]
                else:
                    tp_kind = Parameter.POSITIONAL_OR_KEYWORD  # type: ignore[assignment]

                tp_ann: Any = Parameter.empty
                tp_default: Any = Parameter.empty

                current = self.fetch_token()
                if current and current.match([token.OP, ':'], [token.OP, '=']):
                    if current == [token.OP, ':']:
                        tokens = self.fetch_type_param_spec()
                        tp_ann = self._build_identifier(tokens)

                    if self.current and self.current == [token.OP, '=']:
                        tokens = self.fetch_type_param_spec()
                        tp_default = self._build_identifier(tokens)

                if (
                    tp_kind != Parameter.POSITIONAL_OR_KEYWORD
                    and tp_ann != Parameter.empty
                ):
                    msg = (
                        'type parameter bound or constraint is not allowed '
                        f'for {tp_kind.description} parameters'
                    )
                    raise SyntaxError(msg)

                type_param = (tp_name, tp_kind, tp_default, tp_ann)
                self.type_params.append(type_param)

    def _build_identifier(self, tokens: list[Token]) -> str:
        idents: list[str] = []
        tokens: Iterable[Token] = iter(tokens)  # type: ignore[no-redef]
        # do not format opening brackets
        for tok in tokens:
            if not tok.match([token.OP, '('], [token.OP, '['], [token.OP, '{']):
                # check if the first non-delimiter character is an unpack operator
                is_unpack_operator = tok.match([token.OP, '*'], [token.OP, ['**']])
                idents.append(self._pformat_token(tok, native=is_unpack_operator))
                break
            idents.append(tok.value)

        # check the remaining tokens
        stop = Token(token.ENDMARKER, '', (-1, -1), (-1, -1), '<sentinel>')
        is_unpack_operator = False
        for tok, op, after in _triplewise(chain(tokens, [stop, stop])):
            ident = self._pformat_token(tok, native=is_unpack_operator)
            idents.append(ident)
            # determine if the next token is an unpack operator depending
            # on the left and right hand side of the operator symbol
            is_unpack_operator = op.match([token.OP, '*'], [token.OP, '**']) and not (
                tok.match(
                    token.NAME,
                    token.NUMBER,
                    token.STRING,
                    [token.OP, ')'],
                    [token.OP, ']'],
                    [token.OP, '}'],
                )
                and after.match(
                    token.NAME,
                    token.NUMBER,
                    token.STRING,
                    [token.OP, '('],
                    [token.OP, '['],
                    [token.OP, '{'],
                )
            )

        return ''.join(idents).strip()

    def _pformat_token(self, tok: Token, native: bool = False) -> str:
        if native:
            return tok.value

        if tok.match(token.NEWLINE, token.ENDMARKER):
            return ''

        if tok.match([token.OP, ':'], [token.OP, ','], [token.OP, '#']):
            return f'{tok.value} '

        # Arithmetic operators are allowed because PEP 695 specifies the
        # default type parameter to be *any* expression (so "T1 << T2" is
        # allowed if it makes sense). The caller is responsible to ensure
        # that a multiplication operator ("*") is not to be confused with
        # an unpack operator (which will not be surrounded by spaces).
        #
        # The operators are ordered according to how likely they are to
        # be used and for (possible) future implementations (e.g., "&" for
        # an intersection type).
        if tok.match(
            # Most likely operators to appear
            [token.OP, '='], [token.OP, '|'],
            # Type composition (future compatibility)
            [token.OP, '&'], [token.OP, '^'], [token.OP, '<'], [token.OP, '>'],
            # Unlikely type composition
            [token.OP, '+'], [token.OP, '-'], [token.OP, '*'], [token.OP, '**'],
            # Unlikely operators but included for completeness
            [token.OP, '@'], [token.OP, '/'], [token.OP, '//'], [token.OP, '%'],
            [token.OP, '<<'], [token.OP, '>>'], [token.OP, '>>>'],
            [token.OP, '<='], [token.OP, '>='], [token.OP, '=='], [token.OP, '!='],
        ):  # fmt: skip
            return f' {tok.value} '

        return tok.value


def _parse_type_list(
    tp_list: str,
    env: BuildEnvironment,
    multi_line_parameter_list: bool = False,
    trailing_comma: bool = True,
) -> addnodes.desc_type_parameter_list:
    """Parse a list of type parameters according to PEP 695."""
    type_params = addnodes.desc_type_parameter_list(tp_list)
    type_params['multi_line_parameter_list'] = multi_line_parameter_list
    type_params['multi_line_trailing_comma'] = trailing_comma
    # formal parameter names are interpreted as type parameter names and
    # type annotations are interpreted as type parameter bound or constraints
    parser = _TypeParameterListParser(tp_list)
    parser.parse()
    for tp_name, tp_kind, tp_default, tp_ann in parser.type_params:
        # no positional-only or keyword-only allowed in a type parameters list
        if tp_kind in {Parameter.POSITIONAL_ONLY, Parameter.KEYWORD_ONLY}:
            msg = (
                'positional-only or keyword-only parameters '
                'are prohibited in type parameter lists'
            )
            raise SyntaxError(msg)

        node = addnodes.desc_type_parameter()
        if tp_kind == Parameter.VAR_POSITIONAL:
            node += addnodes.desc_sig_operator('', '*')
        elif tp_kind == Parameter.VAR_KEYWORD:
            node += addnodes.desc_sig_operator('', '**')
        node += addnodes.desc_sig_name('', tp_name)

        if tp_ann is not Parameter.empty:
            annotation = _parse_annotation(tp_ann, env)
            if not annotation:
                continue

            node += addnodes.desc_sig_punctuation('', ':')
            node += addnodes.desc_sig_space()

            type_ann_expr = addnodes.desc_sig_name('', '', *annotation)  # type: ignore[arg-type]
            # a type bound is ``T: U`` whereas type constraints
            # must be enclosed with parentheses. ``T: (U, V)``
            if tp_ann.startswith('(') and tp_ann.endswith(')'):
                type_ann_text = type_ann_expr.astext()
                if type_ann_text.startswith('(') and type_ann_text.endswith(')'):
                    node += type_ann_expr
                else:
                    # surrounding braces are lost when using _parse_annotation()
                    node += addnodes.desc_sig_punctuation('', '(')
                    node += type_ann_expr  # type constraint
                    node += addnodes.desc_sig_punctuation('', ')')
            else:
                node += type_ann_expr  # type bound

        if tp_default is not Parameter.empty:
            # Always surround '=' with spaces, even if there is no annotation
            node += addnodes.desc_sig_space()
            node += addnodes.desc_sig_operator('', '=')
            node += addnodes.desc_sig_space()
            node += nodes.inline(
                '', tp_default, classes=['default_value'], support_smartquotes=False
            )

        type_params += node
    return type_params


def _parse_arglist(
    arglist: str,
    env: BuildEnvironment,
    multi_line_parameter_list: bool = False,
    trailing_comma: bool = True,
) -> addnodes.desc_parameterlist:
    """Parse a list of arguments using AST parser"""
    params = addnodes.desc_parameterlist(arglist)
    params['multi_line_parameter_list'] = multi_line_parameter_list
    params['multi_line_trailing_comma'] = trailing_comma
    sig = signature_from_str('(%s)' % arglist)
    last_kind = None
    for param in sig.parameters.values():
        if param.kind != param.POSITIONAL_ONLY and last_kind == param.POSITIONAL_ONLY:
            params += _positional_only_separator()
        if param.kind == param.KEYWORD_ONLY and last_kind in {
            param.POSITIONAL_OR_KEYWORD,
            param.POSITIONAL_ONLY,
            None,
        }:
            params += _keyword_only_separator()

        node = addnodes.desc_parameter()
        if param.kind == param.VAR_POSITIONAL:
            node += addnodes.desc_sig_operator('', '*')
            node += addnodes.desc_sig_name('', param.name)
        elif param.kind == param.VAR_KEYWORD:
            node += addnodes.desc_sig_operator('', '**')
            node += addnodes.desc_sig_name('', param.name)
        else:
            node += addnodes.desc_sig_name('', param.name)

        if param.annotation is not param.empty:
            children = _parse_annotation(param.annotation, env)
            node += addnodes.desc_sig_punctuation('', ':')
            node += addnodes.desc_sig_space()
            node += addnodes.desc_sig_name('', '', *children)  # type: ignore[arg-type]
        if param.default is not param.empty:
            if param.annotation is not param.empty:
                node += addnodes.desc_sig_space()
                node += addnodes.desc_sig_operator('', '=')
                node += addnodes.desc_sig_space()
            else:
                node += addnodes.desc_sig_operator('', '=')
            node += nodes.inline(
                '', param.default, classes=['default_value'], support_smartquotes=False
            )

        params += node
        last_kind = param.kind

    if last_kind == Parameter.POSITIONAL_ONLY:
        params += _positional_only_separator()

    return params


def _positional_only_separator() -> addnodes.desc_parameter:
    # PEP 570: Separator for positional only parameters: /
    positional_only_abbr = nodes.abbreviation(
        '/', '/', explanation=_('Positional-only parameter separator (PEP 570)')
    )
    positional_only_op = addnodes.desc_sig_operator(
        '/', '', positional_only_abbr, classes=['positional-only-separator']
    )
    return addnodes.desc_parameter('/', '', positional_only_op)


def _keyword_only_separator() -> addnodes.desc_parameter:
    # PEP 3102: Separator for keyword only parameters: *
    keyword_only_abbr = nodes.abbreviation(
        '*', '*', explanation=_('Keyword-only parameters separator (PEP 3102)')
    )
    keyword_only_op = addnodes.desc_sig_operator(
        '*', '', keyword_only_abbr, classes=['keyword-only-separator']
    )
    return addnodes.desc_parameter('*', '', keyword_only_op)


def _pseudo_parse_arglist(
    signode: desc_signature,
    arglist: str,
    *,
    multi_line_parameter_list: bool = False,
    trailing_comma: bool = True,
    env: BuildEnvironment,
) -> None:
    """'Parse' a list of arguments separated by commas.

    Arguments can have "optional" annotations given by enclosing them in
    brackets.  Currently, this will split at any comma, even if it's inside a
    string literal (e.g. default argument value).
    """
    # TODO: decompose 'env' parameter into only the required bits
    paramlist = addnodes.desc_parameterlist()
    paramlist['multi_line_parameter_list'] = multi_line_parameter_list
    paramlist['multi_line_trailing_comma'] = trailing_comma
    stack: list[Element] = [paramlist]
    try:
        for argument in arglist.split(','):
            argument = argument.strip()
            ends_open = ends_close = 0
            while argument.startswith('['):
                stack.append(addnodes.desc_optional())
                stack[-2] += stack[-1]
                argument = argument[1:].strip()
            while argument.startswith(']'):
                stack.pop()
                argument = argument[1:].strip()
            while argument.endswith(']') and not argument.endswith('[]'):
                ends_close += 1
                argument = argument[:-1].strip()
            while argument.endswith('['):
                ends_open += 1
                argument = argument[:-1].strip()
            if argument:
                param_with_annotation, _, default_value = argument.partition('=')
                param_name, _, annotation = param_with_annotation.partition(':')
                del param_with_annotation

                node = addnodes.desc_parameter()
                node += addnodes.desc_sig_name('', param_name.strip())
                if annotation:
                    children = _parse_annotation(annotation.strip(), env=env)
                    node += addnodes.desc_sig_punctuation('', ':')
                    node += addnodes.desc_sig_space()
                    node += addnodes.desc_sig_name('', '', *children)  # type: ignore[arg-type]
                if default_value:
                    if annotation:
                        node += addnodes.desc_sig_space()
                    node += addnodes.desc_sig_operator('', '=')
                    if annotation:
                        node += addnodes.desc_sig_space()
                    node += nodes.inline(
                        '',
                        default_value.strip(),
                        classes=['default_value'],
                        support_smartquotes=False,
                    )
                stack[-1] += node
            while ends_open:
                stack.append(addnodes.desc_optional())
                stack[-2] += stack[-1]
                ends_open -= 1
            while ends_close:
                stack.pop()
                ends_close -= 1
        if len(stack) != 1:
            raise IndexError  # NoQA: TRY301
    except IndexError:
        # if there are too few or too many elements on the stack, just give up
        # and treat the whole argument list as one argument, discarding the
        # already partially populated paramlist node
        paramlist = addnodes.desc_parameterlist()
        paramlist += addnodes.desc_parameter(arglist, arglist)
        signode += paramlist
    else:
        signode += paramlist


def _triplewise(iterable: Iterable[Token]) -> Iterator[tuple[Token, ...]]:
    # sliding_window('ABCDEFG', 4) --> ABCD BCDE CDEF DEFG
    it = iter(iterable)
    window = deque(islice(it, 3), maxlen=3)
    if len(window) == 3:
        yield tuple(window)
    for x in it:
        window.append(x)
        yield tuple(window)



from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx.parsers import RSTParser
from sphinx.util.docutils import _parse_str_to_doctree

if TYPE_CHECKING:
    from docutils import nodes

    from sphinx.application import Sphinx


def parse(app: Sphinx, text: str, docname: str = 'index') -> nodes.document:
    """Parse a string as reStructuredText with Sphinx."""
    config = app.config
    env = app.env
    registry = app.registry
    srcdir = app.srcdir

    # Get settings
    settings_overrides = {
        'env': env,
        'gettext_compact': True,
        'input_encoding': 'utf-8',
        'output_encoding': 'unicode',
        'traceback': True,
    }

    # Create parser
    parser = RSTParser()
    parser._config = config
    parser._env = env

    env.current_document.docname = docname
    try:
        return _parse_str_to_doctree(
            text,
            filename=srcdir / f'{docname}.rst',
            default_settings=settings_overrides,
            env=env,
            parser=parser,
            transforms=registry.get_transforms(),
        )
    finally:
        env.current_document.docname = ''



"""Input/Output files"""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from docutils.io import FileInput
from docutils.readers import standalone
from docutils.transforms.references import DanglingReferences
from docutils.writers import UnfilteredWriter

from sphinx.deprecation import RemovedInSphinx10Warning
from sphinx.transforms import SphinxTransformer
from sphinx.util import logging
from sphinx.util.docutils import LoggingReporter

if TYPE_CHECKING:
    from typing import Any

    from docutils import nodes
    from docutils.io import Input
    from docutils.parsers import Parser
    from docutils.transforms import Transform

    from sphinx.environment import BuildEnvironment
    from sphinx.util.docutils import _DocutilsSettings


logger = logging.getLogger(__name__)

warnings.warn('sphinx.io is deprecated', RemovedInSphinx10Warning, stacklevel=2)


class SphinxBaseReader(standalone.Reader):  # type: ignore[type-arg]
    """A base class of readers for Sphinx.

    This replaces reporter by Sphinx's on generating document.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        warnings.warn(
            'sphinx.io.SphinxBaseReader is deprecated',
            RemovedInSphinx10Warning,
            stacklevel=2,
        )

    transforms: list[type[Transform]] = []

    def get_transforms(self) -> list[type[Transform]]:
        transforms = super().get_transforms() + self.transforms

        # remove transforms which is not needed for Sphinx
        unused = [DanglingReferences]
        for transform in unused:
            if transform in transforms:
                transforms.remove(transform)

        return transforms

    def new_document(self) -> nodes.document:
        """Creates a new document object which has a special reporter object good
        for logging.
        """
        document = super().new_document()

        # substitute transformer
        document.transformer = SphinxTransformer(document)
        document.transformer.set_environment(self.settings.env)

        # substitute reporter
        reporter = document.reporter
        document.reporter = LoggingReporter.from_reporter(reporter)

        return document


class SphinxStandaloneReader(SphinxBaseReader):
    """A basic document reader for Sphinx."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        warnings.warn(
            'sphinx.io.SphinxStandaloneReader is deprecated',
            RemovedInSphinx10Warning,
            stacklevel=2,
        )

    def _setup_transforms(self, transforms: list[type[Transform]], /) -> None:
        self.transforms = self.transforms + transforms

    def read(
        self,
        source: Input,  # type: ignore[type-arg]
        parser: Parser,
        settings: _DocutilsSettings,
    ) -> nodes.document:
        self.source = source
        if not self.parser:
            self.parser = parser
        self.settings = settings
        self.input = self.read_source(settings.env)
        self.parse()
        assert self.document is not None
        return self.document

    def read_source(self, env: BuildEnvironment) -> str:
        """Read content from source and do post-process."""
        assert self.source is not None
        content = self.source.read()

        # emit "source-read" event
        arg = [content]
        env.events.emit('source-read', env.current_document.docname, arg)
        return arg[0]


class SphinxDummyWriter(UnfilteredWriter):  # type: ignore[type-arg]
    """Dummy writer module used for generating doctree."""

    def __init__(self) -> None:
        super().__init__()
        warnings.warn(
            'sphinx.io.SphinxDummyWriter is deprecated',
            RemovedInSphinx10Warning,
            stacklevel=2,
        )

    supported = ('html',)  # needed to keep "meta" nodes

    def translate(self) -> None:
        pass


def SphinxDummySourceClass(source: Any, *args: Any, **kwargs: Any) -> Any:
    """Bypass source object as is to cheat Publisher."""
    warnings.warn(
        'sphinx.io.SphinxDummySourceClass is deprecated',
        RemovedInSphinx10Warning,
        stacklevel=2,
    )
    return source


class SphinxFileInput(FileInput):
    """A basic FileInput for Sphinx."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs['error_handler'] = 'sphinx'
        super().__init__(*args, **kwargs)
        warnings.warn(
            'sphinx.io.SphinxFileInput is deprecated',
            RemovedInSphinx10Warning,
            stacklevel=2,
        )



# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/portuguese/stop.txt

from __future__ import annotations

PORTUGUESE_STOPWORDS = frozenset({
    'a',
    'ao',
    'aos',
    'aquela',
    'aquelas',
    'aquele',
    'aqueles',
    'aquilo',
    'as',
    'até',
    'com',
    'como',
    'da',
    'das',
    'de',
    'dela',
    'delas',
    'dele',
    'deles',
    'depois',
    'do',
    'dos',
    'e',
    'ela',
    'elas',
    'ele',
    'eles',
    'em',
    'entre',
    'era',
    'eram',
    'essa',
    'essas',
    'esse',
    'esses',
    'esta',
    'estamos',
    'estas',
    'estava',
    'estavam',
    'este',
    'esteja',
    'estejam',
    'estejamos',
    'estes',
    'esteve',
    'estive',
    'estivemos',
    'estiver',
    'estivera',
    'estiveram',
    'estiverem',
    'estivermos',
    'estivesse',
    'estivessem',
    'estivéramos',
    'estivéssemos',
    'estou',
    'está',
    'estávamos',
    'estão',
    'eu',
    'foi',
    'fomos',
    'for',
    'fora',
    'foram',
    'forem',
    'formos',
    'fosse',
    'fossem',
    'fui',
    'fôramos',
    'fôssemos',
    'haja',
    'hajam',
    'hajamos',
    'havemos',
    'hei',
    'houve',
    'houvemos',
    'houver',
    'houvera',
    'houveram',
    'houverei',
    'houverem',
    'houveremos',
    'houveria',
    'houveriam',
    'houvermos',
    'houverá',
    'houverão',
    'houveríamos',
    'houvesse',
    'houvessem',
    'houvéramos',
    'houvéssemos',
    'há',
    'hão',
    'isso',
    'isto',
    'já',
    'lhe',
    'lhes',
    'mais',
    'mas',
    'me',
    'mesmo',
    'meu',
    'meus',
    'minha',
    'minhas',
    'muito',
    'na',
    'nas',
    'nem',
    'no',
    'nos',
    'nossa',
    'nossas',
    'nosso',
    'nossos',
    'num',
    'numa',
    'não',
    'nós',
    'o',
    'os',
    'ou',
    'para',
    'pela',
    'pelas',
    'pelo',
    'pelos',
    'por',
    'qual',
    'quando',
    'que',
    'quem',
    'se',
    'seja',
    'sejam',
    'sejamos',
    'sem',
    'serei',
    'seremos',
    'seria',
    'seriam',
    'será',
    'serão',
    'seríamos',
    'seu',
    'seus',
    'somos',
    'sou',
    'sua',
    'suas',
    'são',
    'só',
    'também',
    'te',
    'tem',
    'temos',
    'tenha',
    'tenham',
    'tenhamos',
    'tenho',
    'terei',
    'teremos',
    'teria',
    'teriam',
    'terá',
    'terão',
    'teríamos',
    'teu',
    'teus',
    'teve',
    'tinha',
    'tinham',
    'tive',
    'tivemos',
    'tiver',
    'tivera',
    'tiveram',
    'tiverem',
    'tivermos',
    'tivesse',
    'tivessem',
    'tivéramos',
    'tivéssemos',
    'tu',
    'tua',
    'tuas',
    'tém',
    'tínhamos',
    'um',
    'uma',
    'você',
    'vocês',
    'vos',
    'à',
    'às',
    'éramos',
})



"""Build epub3 files.

Originally derived from epub.py.
"""

from __future__ import annotations

import html
import os
import re
import time
from typing import TYPE_CHECKING, NamedTuple

from sphinx import package_dir
from sphinx.builders import _epub_base
from sphinx.config import ENUM
from sphinx.locale import __
from sphinx.util import logging
from sphinx.util.fileutil import copy_asset_file
from sphinx.util.osutil import make_filename

if TYPE_CHECKING:
    from collections.abc import Set
    from typing import Any

    from sphinx.application import Sphinx
    from sphinx.config import Config
    from sphinx.util.typing import ExtensionMetadata

logger = logging.getLogger(__name__)


class NavPoint(NamedTuple):
    text: str
    refuri: str
    children: list[NavPoint]


# writing modes
PAGE_PROGRESSION_DIRECTIONS = {
    'horizontal': 'ltr',
    'vertical': 'rtl',
}
IBOOK_SCROLL_AXIS = {
    'horizontal': 'vertical',
    'vertical': 'horizontal',
}
THEME_WRITING_MODES = {
    'vertical': 'vertical-rl',
    'horizontal': 'horizontal-tb',
}

DOCTYPE = """<!DOCTYPE html>"""

HTML_TAG = (
    '<html xmlns="http://www.w3.org/1999/xhtml" '
    'xmlns:epub="http://www.idpf.org/2007/ops">'
)

# https://www.w3.org/TR/REC-xml/#NT-Name
_xml_name_start_char = (
    ':|[A-Z]|_|[a-z]|[\u00c0-\u00d6]'
    '|[\u00d8-\u00f6]|[\u00f8-\u02ff]|[\u0370-\u037d]'
    '|[\u037f-\u1fff]|[\u200c-\u200d]|[\u2070-\u218f]'
    '|[\u2c00-\u2fef]|[\u3001-\ud7ff]|[\uf900-\ufdcf]'
    '|[\ufdf0-\ufffd]|[\U00010000-\U000effff]'
)
_xml_name_char = (
    _xml_name_start_char + r'\-|\.|[0-9]|\u00b7|[\u0300-\u036f]|[\u203f-\u2040]'
)
_XML_NAME_PATTERN = re.compile(f'({_xml_name_start_char})({_xml_name_char})*')


class Epub3Builder(_epub_base.EpubBuilder):
    """Builder that outputs epub3 files.

    It creates the metainfo files content.opf, nav.xhtml, toc.ncx, mimetype,
    and META-INF/container.xml. Afterwards, all necessary files are zipped to
    an epub file.
    """

    name = 'epub'
    epilog = __('The ePub file is in %(outdir)s.')

    supported_remote_images = False
    template_dir = package_dir.joinpath('templates', 'epub3')
    doctype = DOCTYPE
    html_tag = HTML_TAG
    use_meta_charset = True

    # Finish by building the epub file
    def handle_finish(self) -> None:
        """Create the metainfo files and finally the epub."""
        self.get_toc()
        self.build_mimetype()
        self.build_container()
        self.build_content()
        self.build_navigation_doc()
        self.build_toc()
        self.build_epub()

    def content_metadata(self) -> dict[str, Any]:
        """Create a dictionary with all metadata for the content.opf
        file properly escaped.
        """
        writing_mode = self.config.epub_writing_mode

        if (source_date_epoch := os.getenv('SOURCE_DATE_EPOCH')) is not None:
            time_tuple = time.gmtime(int(source_date_epoch))
        else:
            time_tuple = time.gmtime()

        metadata = super().content_metadata()
        metadata['description'] = html.escape(self.config.epub_description)
        metadata['contributor'] = html.escape(self.config.epub_contributor)
        metadata['page_progression_direction'] = PAGE_PROGRESSION_DIRECTIONS.get(
            writing_mode
        )
        metadata['ibook_scroll_axis'] = IBOOK_SCROLL_AXIS.get(writing_mode)
        metadata['date'] = html.escape(time.strftime('%Y-%m-%dT%H:%M:%SZ', time_tuple))
        metadata['version'] = html.escape(self.config.version)
        metadata['epub_version'] = self.config.epub_version
        return metadata

    def prepare_writing(self, docnames: Set[str]) -> None:
        super().prepare_writing(docnames)

        writing_mode = self.config.epub_writing_mode
        self.globalcontext['theme_writing_mode'] = THEME_WRITING_MODES.get(writing_mode)
        self.globalcontext['html_tag'] = self.html_tag
        self.globalcontext['use_meta_charset'] = self.use_meta_charset
        self.globalcontext['skip_ua_compatible'] = True

    def build_navlist(self, navnodes: list[dict[str, Any]]) -> list[NavPoint]:
        """Create the toc navigation structure.

        This method is almost same as build_navpoints method in epub.py.
        This is because the logical navigation structure of epub3 is not
        different from one of epub2.

        The difference from build_navpoints method is templates which are used
        when generating navigation documents.
        """
        navstack: list[NavPoint] = [NavPoint('', '', [])]
        level = 0
        for node in navnodes:
            if not node['text']:
                continue
            file = node['refuri'].split('#')[0]
            if file in self.ignored_files:
                continue
            if node['level'] > self.config.epub_tocdepth:
                continue

            navpoint = NavPoint(node['text'], node['refuri'], [])
            if node['level'] == level:
                navstack.pop()
                navstack[-1].children.append(navpoint)
                navstack.append(navpoint)
            elif node['level'] == level + 1:
                level += 1
                navstack[-1].children.append(navpoint)
                navstack.append(navpoint)
            elif node['level'] < level:
                while node['level'] < len(navstack):
                    navstack.pop()
                level = node['level']
                navstack[-1].children.append(navpoint)
                navstack.append(navpoint)
            else:
                unreachable = 'Should never reach here. It might be a bug.'
                raise RuntimeError(unreachable)

        return navstack[0].children

    def navigation_doc_metadata(self, navlist: list[NavPoint]) -> dict[str, Any]:
        """Create a dictionary with all metadata for the nav.xhtml file
        properly escaped.
        """
        return {
            'lang': html.escape(self.config.epub_language),
            'toc_locale': html.escape(self.guide_titles['toc']),
            'navlist': navlist,
        }

    def build_navigation_doc(self) -> None:
        """Write the metainfo file nav.xhtml."""
        logger.info(__('writing nav.xhtml file...'))

        if self.config.epub_tocscope == 'default':
            doctree = self.env.get_and_resolve_doctree(
                self.config.root_doc,
                self,
                tags=self.tags,
                prune_toctrees=False,
                includehidden=False,
            )
            refnodes = self.get_refnodes(doctree, [])
            self.toc_add_files(refnodes)
        else:
            # 'includehidden'
            refnodes = self.refnodes
        navlist = self.build_navlist(refnodes)
        copy_asset_file(
            self.template_dir / 'nav.xhtml.jinja',
            self.outdir,
            context=self.navigation_doc_metadata(navlist),
            force=True,
        )

        # Add nav.xhtml to epub file
        if 'nav.xhtml' not in self.files:
            self.files.append('nav.xhtml')


def validate_config_values(app: Sphinx) -> None:
    if app.builder.name != 'epub':
        return

    # <package> lang attribute, dc:language
    if not app.config.epub_language:
        logger.warning(
            __(
                'conf value "epub_language" (or "language") '
                'should not be empty for EPUB3'
            )
        )
    # <package> unique-identifier attribute
    if not _XML_NAME_PATTERN.match(app.config.epub_uid):
        logger.warning(__('conf value "epub_uid" should be XML NAME for EPUB3'))
    # dc:title
    if not app.config.epub_title:
        logger.warning(
            __(
                'conf value "epub_title" (or "html_title") '
                'should not be empty for EPUB3'
            )
        )
    # dc:creator
    if not app.config.epub_author:
        logger.warning(__('conf value "epub_author" should not be empty for EPUB3'))
    # dc:contributor
    if not app.config.epub_contributor:
        logger.warning(
            __('conf value "epub_contributor" should not be empty for EPUB3')
        )
    # dc:description
    if not app.config.epub_description:
        logger.warning(
            __('conf value "epub_description" should not be empty for EPUB3')
        )
    # dc:publisher
    if not app.config.epub_publisher:
        logger.warning(__('conf value "epub_publisher" should not be empty for EPUB3'))
    # dc:rights
    if not app.config.epub_copyright:
        logger.warning(
            __(
                'conf value "epub_copyright" (or "copyright")'
                'should not be empty for EPUB3'
            )
        )
    # dc:identifier
    if not app.config.epub_identifier:
        logger.warning(__('conf value "epub_identifier" should not be empty for EPUB3'))
    # meta ibooks:version
    if not app.config.version:
        logger.warning(__('conf value "version" should not be empty for EPUB3'))


def convert_epub_css_files(app: Sphinx, config: Config) -> None:
    """Convert string styled epub_css_files to tuple styled one."""
    epub_css_files: list[tuple[str, dict[str, Any]]] = []
    for entry in config.epub_css_files:
        if isinstance(entry, str):
            epub_css_files.append((entry, {}))
        else:
            try:
                filename, attrs = entry
                epub_css_files.append((filename, attrs))
            except Exception:
                logger.warning(__('invalid css_file: %r, ignored'), entry)
                continue

    config.epub_css_files = epub_css_files


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_builder(Epub3Builder)

    # config values
    app.add_config_value(
        'epub_basename',
        lambda self: make_filename(self.project),
        '',
        types=frozenset({str}),
    )
    app.add_config_value(
        'epub_version', 3.0, 'epub', types=frozenset({float})
    )  # experimental
    app.add_config_value('epub_theme', 'epub', 'epub', types=frozenset({str}))
    app.add_config_value('epub_theme_options', {}, 'epub', types=frozenset({dict}))
    app.add_config_value(
        'epub_title', lambda self: self.project, 'epub', types=frozenset({str})
    )
    app.add_config_value(
        'epub_author', lambda self: self.author, 'epub', types=frozenset({str})
    )
    app.add_config_value(
        'epub_language',
        lambda self: self.language or 'en',
        'epub',
        types=frozenset({str}),
    )
    app.add_config_value(
        'epub_publisher', lambda self: self.author, 'epub', types=frozenset({str})
    )
    app.add_config_value(
        'epub_copyright', lambda self: self.copyright, 'epub', types=frozenset({str})
    )
    app.add_config_value('epub_identifier', 'unknown', 'epub', types=frozenset({str}))
    app.add_config_value('epub_scheme', 'unknown', 'epub', types=frozenset({str}))
    app.add_config_value('epub_uid', 'unknown', 'env', types=frozenset({str}))
    app.add_config_value('epub_cover', (), 'env', types=frozenset({list, tuple}))
    app.add_config_value('epub_guide', (), 'env', types=frozenset({list, tuple}))
    app.add_config_value('epub_pre_files', [], 'env', types=frozenset({list, tuple}))
    app.add_config_value('epub_post_files', [], 'env', types=frozenset({list, tuple}))
    app.add_config_value(
        'epub_css_files',
        lambda config: config.html_css_files,
        'epub',
        types=frozenset({list, tuple}),
    )
    app.add_config_value(
        'epub_exclude_files', [], 'env', types=frozenset({list, tuple})
    )
    app.add_config_value('epub_tocdepth', 3, 'env', types=frozenset({int}))
    app.add_config_value('epub_tocdup', True, 'env', types=frozenset({bool}))
    app.add_config_value('epub_tocscope', 'default', 'env', types=frozenset({str}))
    app.add_config_value('epub_fix_images', False, 'env', types=frozenset({bool}))
    app.add_config_value('epub_max_image_width', 0, 'env', types=frozenset({int}))
    app.add_config_value('epub_show_urls', 'inline', 'epub', types=frozenset({str}))
    app.add_config_value(
        'epub_use_index',
        lambda self: self.html_use_index,
        'epub',
        types=frozenset({bool}),
    )
    app.add_config_value('epub_description', 'unknown', 'epub', types=frozenset({str}))
    app.add_config_value('epub_contributor', 'unknown', 'epub', types=frozenset({str}))
    app.add_config_value(
        'epub_writing_mode', 'horizontal', 'epub', types=ENUM('horizontal', 'vertical')
    )

    # event handlers
    app.connect('config-inited', convert_epub_css_files, priority=800)
    app.connect('builder-inited', validate_config_values)

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx.domains.c._ast import (
    ASTDeclaration,
    ASTNestedName,
)
from sphinx.locale import __
from sphinx.util import logging

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Iterator, Sequence
    from typing import Any, Self

    from sphinx.domains.c._ast import (
        ASTIdentifier,
    )
    from sphinx.environment import BuildEnvironment

logger = logging.getLogger(__name__)


class _DuplicateSymbolError(Exception):
    def __init__(self, symbol: Symbol, declaration: ASTDeclaration) -> None:
        assert symbol
        assert declaration
        self.symbol = symbol
        self.declaration = declaration

    def __str__(self) -> str:
        return 'Internal C duplicate symbol error:\n%s' % self.symbol.dump(0)


class SymbolLookupResult:
    __slots__ = 'symbols', 'parent_symbol', 'ident'

    symbols: Iterable[Symbol]
    parent_symbol: Symbol
    ident_or_op: ASTIdentifier

    def __init__(
        self, symbols: Iterable[Symbol], parent_symbol: Symbol, ident: ASTIdentifier
    ) -> None:
        self.symbols = symbols
        self.parent_symbol = parent_symbol
        self.ident = ident

    @property
    def parentSymbol(self) -> Symbol:
        return self.parent_symbol


class LookupKey:
    __slots__ = ('data',)

    data: Sequence[tuple[ASTIdentifier, str]]

    def __init__(self, data: Sequence[tuple[ASTIdentifier, str]], /) -> None:
        self.data = data

    def __repr__(self) -> str:
        return f'LookupKey({self.data!r})'

    def __str__(self) -> str:
        inner = ', '.join(f'({ident}, {id_})' for ident, id_ in self.data)
        return f'[{inner}]'


class Symbol:
    debug_indent = 0
    debug_indent_string = '  '
    debug_lookup = False
    debug_show_tree = False

    def __copy__(self) -> Self:
        raise AssertionError  # shouldn't happen

    def __deepcopy__(self, memo: Any) -> Symbol:
        if self.parent:
            raise AssertionError  # shouldn't happen
        # the domain base class makes a copy of the initial data, which is fine
        return Symbol(None, None, None, None, None)

    @staticmethod
    def debug_print(*args: Any) -> None:
        msg = Symbol.debug_indent_string * Symbol.debug_indent
        msg += ''.join(str(e) for e in args)
        logger.debug(msg)

    def _assert_invariants(self) -> None:
        if not self.parent:
            # parent == None means global scope, so declaration means a parent
            assert not self.declaration
            assert not self.docname
        else:
            if self.declaration:
                assert self.docname

    def __setattr__(self, key: str, value: Any) -> None:
        if key == 'children':
            raise AssertionError
        return super().__setattr__(key, value)

    def __init__(
        self,
        parent: Symbol | None,
        ident: ASTIdentifier | None,
        declaration: ASTDeclaration | None,
        docname: str | None,
        line: int | None,
    ) -> None:
        self.parent = parent
        # declarations in a single directive are linked together
        self.siblingAbove: Symbol | None = None
        self.siblingBelow: Symbol | None = None
        self.ident = ident
        self.declaration = declaration
        self.docname = docname
        self.line = line
        self.isRedeclaration = False
        self._assert_invariants()

        # These properties store the same children for different access patterns.
        # ``_add_child()`` and ``_remove_child()`` should be used for modifying them.
        self._children_by_name: dict[str, Symbol] = {}
        self._children_by_docname: dict[str, dict[str, Symbol]] = {}
        self._anon_children: set[Symbol] = set()

        if self.parent:
            self.parent._add_child(self)
        if self.declaration:
            self.declaration.symbol = self

        # Do symbol addition after self._children has been initialised.
        self._add_function_params()

    def __repr__(self) -> str:
        return f'<Symbol {self.to_string(indent=0)!r}>'

    @property
    def _children(self) -> Iterable[Symbol]:
        return self._children_by_name.values()

    def _add_child(self, child: Symbol) -> None:
        name = child.ident.name
        if name in self._children_by_name:
            # Duplicate so don't add - will be reported in _add_symbols()
            return
        self._children_by_name[name] = child
        child_docname: str = child.docname
        self._children_by_docname.setdefault(child_docname, {})[name] = child
        if child.ident.is_anonymous:
            self._anon_children.add(child)

    def _remove_child(self, child: Symbol) -> None:
        name = child.ident.name
        self._children_by_name.pop(name, None)
        if children := self._children_by_docname.get(child.docname):
            children.pop(name, None)
        if child.ident.is_anonymous:
            self._anon_children.discard(child)

    def _fill_empty(self, declaration: ASTDeclaration, docname: str, line: int) -> None:
        self._assert_invariants()
        assert self.declaration is None
        assert self.docname is None
        assert self.line is None
        assert declaration is not None
        assert docname is not None
        assert line is not None
        self.declaration = declaration
        self.declaration.symbol = self
        self.docname = docname
        self.line = line
        self._assert_invariants()
        # and symbol addition should be done as well
        self._add_function_params()

    def _add_function_params(self) -> None:
        if Symbol.debug_lookup:
            Symbol.debug_indent += 1
            Symbol.debug_print('_add_function_params:')
        # Note: we may be called from _fill_empty, so the symbols we want
        #       to add may actually already be present (as empty symbols).

        # add symbols for function parameters, if any
        if (
            self.declaration is not None
            and self.declaration.function_params is not None
        ):
            for p in self.declaration.function_params:
                if p.arg is None:
                    continue
                nn = p.arg.name
                if nn is None:
                    continue
                # (comparing to the template params: we have checked that we are a declaration)
                decl = ASTDeclaration('functionParam', None, p)
                assert not nn.rooted
                assert len(nn.names) == 1
                self._add_symbols(nn, decl, self.docname, self.line)
        if Symbol.debug_lookup:
            Symbol.debug_indent -= 1

    def remove(self) -> None:
        if self.parent:
            self.parent._remove_child(self)
            self.parent = None

    def clear_doc(self, docname: str) -> None:
        if docname not in self._children_by_docname:
            for child in self._children:
                child.clear_doc(docname)
            return

        children: dict[str, Symbol] = self._children_by_docname.pop(docname)
        for child in children.values():
            child.declaration = None
            child.docname = None
            child.line = None
            if child.siblingAbove is not None:
                child.siblingAbove.siblingBelow = child.siblingBelow
            if child.siblingBelow is not None:
                child.siblingBelow.siblingAbove = child.siblingAbove
            child.siblingAbove = None
            child.siblingBelow = None
            self._remove_child(child)

    def get_all_symbols(self) -> Iterator[Symbol]:
        yield self
        for s_child in self._children:
            yield from s_child.get_all_symbols()

    @property
    def children(self) -> Iterator[Symbol]:
        yield from self._children

    def get_lookup_key(self) -> LookupKey:
        # The pickle files for the environment and for each document are distinct.
        # The environment has all the symbols, but the documents has xrefs that
        # must know their scope. A lookup key is essentially a specification of
        # how to find a specific symbol.
        symbols = []
        s = self
        while s.parent:
            symbols.append(s)
            s = s.parent
        key = [
            # TODO: do we need the ID?
            (s.ident, None if s.declaration is None else s.declaration.get_newest_id())
            for s in reversed(symbols)
        ]
        return LookupKey(key)

    def get_full_nested_name(self) -> ASTNestedName:
        symbols = []
        s = self
        while s.parent:
            symbols.append(s)
            s = s.parent
        symbols.reverse()
        names = [s.ident for s in symbols]
        return ASTNestedName(names, rooted=False)

    def _symbol_lookup(
        self,
        nested_name: ASTNestedName,
        on_missing_qualified_symbol: Callable[[Symbol, ASTIdentifier], Symbol | None],
        ancestor_lookup_type: str | None,
        match_self: bool,
        recurse_in_anon: bool,
        search_in_siblings: bool,
    ) -> SymbolLookupResult | None:
        # TODO: further simplification from C++ to C
        # ancestor_lookup_type: if not None, specifies the target type of the lookup
        if Symbol.debug_lookup:
            Symbol.debug_indent += 1
            Symbol.debug_print('_symbol_lookup:')
            Symbol.debug_indent += 1
            Symbol.debug_print('self:')
            logger.debug(self.to_string(Symbol.debug_indent + 1, addEndNewline=False))
            Symbol.debug_print('nested_name:         ', nested_name)
            Symbol.debug_print('ancestor_lookup_type:', ancestor_lookup_type)
            Symbol.debug_print('match_self:          ', match_self)
            Symbol.debug_print('recurse_in_anon:     ', recurse_in_anon)
            Symbol.debug_print('search_in_siblings:  ', search_in_siblings)

        names = nested_name.names

        # find the right starting point for lookup
        parent_symbol = self
        if nested_name.rooted:
            while parent_symbol.parent is not None:
                parent_symbol = parent_symbol.parent

        if ancestor_lookup_type is not None:
            # walk up until we find the first identifier
            first_name = names[0]
            while parent_symbol.parent:
                if first_name.name in parent_symbol._children_by_name:
                    break
                parent_symbol = parent_symbol.parent

        if Symbol.debug_lookup:
            Symbol.debug_print('starting point:')
            logger.debug(
                parent_symbol.to_string(Symbol.debug_indent + 1, addEndNewline=False)
            )

        # and now the actual lookup
        for ident in names[:-1]:
            name = ident.name
            if name in parent_symbol._children_by_name:
                symbol = parent_symbol._children_by_name[name]
            else:
                symbol = on_missing_qualified_symbol(parent_symbol, ident)
                if symbol is None:
                    if Symbol.debug_lookup:
                        Symbol.debug_indent -= 2
                    return None
            parent_symbol = symbol

        if Symbol.debug_lookup:
            Symbol.debug_print('handle last name from:')
            logger.debug(
                parent_symbol.to_string(Symbol.debug_indent + 1, addEndNewline=False)
            )

        # handle the last name
        ident = names[-1]
        name = ident.name
        symbol = parent_symbol._children_by_name.get(name)
        if not symbol and recurse_in_anon:
            for child in parent_symbol._anon_children:
                if name in child._children_by_name:
                    symbol = child._children_by_name[name]
                    break

        if Symbol.debug_lookup:
            Symbol.debug_indent -= 2

        result = [symbol] if symbol else []
        return SymbolLookupResult(result, parent_symbol, ident)

    def _add_symbols(
        self,
        nested_name: ASTNestedName,
        declaration: ASTDeclaration | None,
        docname: str | None,
        line: int | None,
    ) -> Symbol:
        # TODO: further simplification from C++ to C
        # Used for adding a whole path of symbols, where the last may or may not
        # be an actual declaration.

        if Symbol.debug_lookup:
            Symbol.debug_indent += 1
            Symbol.debug_print('_add_symbols:')
            Symbol.debug_indent += 1
            Symbol.debug_print('nn:       ', nested_name)
            Symbol.debug_print('decl:     ', declaration)
            Symbol.debug_print(f'location: {docname}:{line}')

        def on_missing_qualified_symbol(
            parent_symbol: Symbol, ident: ASTIdentifier
        ) -> Symbol:
            if Symbol.debug_lookup:
                Symbol.debug_indent += 1
                Symbol.debug_print('_add_symbols, on_missing_qualified_symbol:')
                Symbol.debug_indent += 1
                Symbol.debug_print('ident: ', ident)
                Symbol.debug_indent -= 2
            return Symbol(
                parent=parent_symbol,
                ident=ident,
                declaration=None,
                docname=None,
                line=None,
            )

        lookup_result = self._symbol_lookup(
            nested_name,
            on_missing_qualified_symbol,
            ancestor_lookup_type=None,
            match_self=False,
            recurse_in_anon=False,
            search_in_siblings=False,
        )
        # we create symbols all the way, so that can't happen
        assert lookup_result is not None
        symbols = list(lookup_result.symbols)
        if len(symbols) == 0:
            if Symbol.debug_lookup:
                Symbol.debug_print('_add_symbols, result, no symbol:')
                Symbol.debug_indent += 1
                Symbol.debug_print('ident:       ', lookup_result.ident)
                Symbol.debug_print('declaration: ', declaration)
                Symbol.debug_print(f'location:    {docname}:{line}')
                Symbol.debug_indent -= 1
            symbol = Symbol(
                parent=lookup_result.parent_symbol,
                ident=lookup_result.ident,
                declaration=declaration,
                docname=docname,
                line=line,
            )
            if Symbol.debug_lookup:
                Symbol.debug_indent -= 2
            return symbol

        if Symbol.debug_lookup:
            Symbol.debug_print('_add_symbols, result, symbols:')
            Symbol.debug_indent += 1
            Symbol.debug_print('number symbols:', len(symbols))
            Symbol.debug_indent -= 1

        if not declaration:
            if Symbol.debug_lookup:
                Symbol.debug_print('no declaration')
                Symbol.debug_indent -= 2
            # good, just a scope creation
            # TODO: what if we have more than one symbol?
            return symbols[0]

        no_decl = []
        with_decl = []
        dup_decl = []
        for s in symbols:
            if s.declaration is None:
                no_decl.append(s)
            elif s.isRedeclaration:
                dup_decl.append(s)
            else:
                with_decl.append(s)
        if Symbol.debug_lookup:
            Symbol.debug_print('#no_decl:  ', len(no_decl))
            Symbol.debug_print('#with_decl:', len(with_decl))
            Symbol.debug_print('#dup_decl: ', len(dup_decl))

        # With partial builds we may start with a large symbol tree stripped of declarations.
        # Essentially any combination of no_decl, with_decl, and dup_decls seems possible.
        # TODO: make partial builds fully work. What should happen when the primary symbol gets
        #  deleted, and other duplicates exist? The full document should probably be rebuild.

        # First check if one of those with a declaration matches.
        # If it's a function, we need to compare IDs,
        # otherwise there should be only one symbol with a declaration.

        if len(with_decl) == 0:
            cand_symbol = None
        else:
            cand_symbol = self._make_cand_symbol(
                lookup_result, declaration, docname, line
            )

            if declaration.objectType != 'function':
                assert len(with_decl) <= 1
                self._handle_duplicate_declaration(
                    with_decl[0], cand_symbol, declaration
                )
                # (not reachable)

            # a function, so compare IDs
            cand_id = declaration.get_newest_id()
            if Symbol.debug_lookup:
                Symbol.debug_print('cand_id:', cand_id)
            for symbol in with_decl:
                old_id = symbol.declaration.get_newest_id()
                if Symbol.debug_lookup:
                    Symbol.debug_print('old_id: ', old_id)
                if cand_id == old_id:
                    self._handle_duplicate_declaration(symbol, cand_symbol, declaration)
                    # (not reachable)
            # no candidate symbol found with matching ID
        # if there is an empty symbol, fill that one
        if len(no_decl) == 0:
            if Symbol.debug_lookup:
                Symbol.debug_print(
                    'no match, no empty, cand_sybmol is not None?:',
                    cand_symbol is not None,
                )
                Symbol.debug_indent -= 2
            if cand_symbol is not None:
                return cand_symbol
            else:
                return self._make_cand_symbol(lookup_result, declaration, docname, line)
        else:
            if Symbol.debug_lookup:
                Symbol.debug_print(
                    'no match, but fill an empty declaration, cand_sybmol is not None?:',
                    cand_symbol is not None,
                )
                Symbol.debug_indent -= 2
            if cand_symbol is not None:
                cand_symbol.remove()
            # assert len(no_decl) == 1
            # TODO: enable assertion when we at some point find out how to do cleanup
            # for now, just take the first one, it should work fine ... right?
            symbol = no_decl[0]
            # If someone first opened the scope, and then later
            # declares it, e.g,
            # .. namespace:: Test
            # .. namespace:: nullptr
            # .. class:: Test
            symbol._fill_empty(declaration, docname, line)
            return symbol

    @staticmethod
    def _make_cand_symbol(
        lookup_result: SymbolLookupResult,
        declaration: ASTDeclaration | None,
        docname: str | None,
        line: int | None,
    ) -> Symbol:
        if Symbol.debug_lookup:
            Symbol.debug_print('begin: creating candidate symbol')
        symbol = Symbol(
            parent=lookup_result.parent_symbol,
            ident=lookup_result.ident,
            declaration=declaration,
            docname=docname,
            line=line,
        )
        if Symbol.debug_lookup:
            Symbol.debug_print('end:   creating candidate symbol')
        return symbol

    @staticmethod
    def _handle_duplicate_declaration(
        symbol: Symbol, cand_symbol: Symbol, declaration: ASTDeclaration
    ) -> None:
        if Symbol.debug_lookup:
            Symbol.debug_indent += 1
            Symbol.debug_print('redeclaration')
            Symbol.debug_indent -= 1
            Symbol.debug_indent -= 2
        # Redeclaration of the same symbol.
        # Let the new one be there, but raise an error to the client
        # so it can use the real symbol as subscope.
        # This will probably result in a duplicate id warning.
        cand_symbol.isRedeclaration = True
        raise _DuplicateSymbolError(symbol, declaration)

    def merge_with(
        self, other: Symbol, docnames: list[str], env: BuildEnvironment
    ) -> None:
        if Symbol.debug_lookup:
            Symbol.debug_indent += 1
            Symbol.debug_print('merge_with:')

        assert other is not None
        for other_child in other._children:
            other_name = other_child.ident.name
            if other_name not in self._children_by_name:
                # TODO: hmm, should we prune by docnames?
                other_child.parent = self
                self._add_child(other_child)
                other_child._assert_invariants()
                continue
            our_child = self._children_by_name[other_name]
            if other_child.declaration and other_child.docname in docnames:
                if not our_child.declaration:
                    our_child._fill_empty(
                        other_child.declaration, other_child.docname, other_child.line
                    )
                elif our_child.docname != other_child.docname:
                    name = str(our_child.declaration)
                    msg = __(
                        'Duplicate C declaration, also defined at %s:%s.\n'
                        "Declaration is '.. c:%s:: %s'."
                    )
                    logger.warning(
                        msg,
                        our_child.docname,
                        our_child.line,
                        our_child.declaration.directiveType,
                        name,
                        location=(other_child.docname, other_child.line),
                        type='duplicate_declaration',
                        subtype='c',
                    )
                else:
                    # Both have declarations, and in the same docname.
                    # This can apparently happen, it should be safe to
                    # just ignore it, right?
                    pass
            our_child.merge_with(other_child, docnames, env)

        if Symbol.debug_lookup:
            Symbol.debug_indent -= 1

    def add_name(self, nestedName: ASTNestedName) -> Symbol:
        if Symbol.debug_lookup:
            Symbol.debug_indent += 1
            Symbol.debug_print('add_name:')
        res = self._add_symbols(nestedName, declaration=None, docname=None, line=None)
        if Symbol.debug_lookup:
            Symbol.debug_indent -= 1
        return res

    def add_declaration(
        self, declaration: ASTDeclaration, docname: str, line: int
    ) -> Symbol:
        if Symbol.debug_lookup:
            Symbol.debug_indent += 1
            Symbol.debug_print('add_declaration:')
        assert declaration is not None
        assert docname is not None
        assert line is not None
        nested_name = declaration.name
        res = self._add_symbols(nested_name, declaration, docname, line)
        if Symbol.debug_lookup:
            Symbol.debug_indent -= 1
        return res

    def find_identifier(
        self,
        ident: ASTIdentifier,
        matchSelf: bool,
        recurseInAnon: bool,
        searchInSiblings: bool,
    ) -> Symbol | None:
        if Symbol.debug_lookup:
            Symbol.debug_indent += 1
            Symbol.debug_print('find_identifier:')
            Symbol.debug_indent += 1
            Symbol.debug_print('ident:           ', ident)
            Symbol.debug_print('matchSelf:       ', matchSelf)
            Symbol.debug_print('recurseInAnon:   ', recurseInAnon)
            Symbol.debug_print('searchInSiblings:', searchInSiblings)
            logger.debug(self.to_string(Symbol.debug_indent + 1, addEndNewline=False))
            Symbol.debug_indent -= 2
        current = self
        while current is not None:
            if Symbol.debug_lookup:
                Symbol.debug_indent += 2
                Symbol.debug_print('trying:')
                logger.debug(
                    current.to_string(Symbol.debug_indent + 1, addEndNewline=False)
                )
                Symbol.debug_indent -= 2
            if matchSelf and current.ident == ident:
                return current
            name = ident.name
            if name in current._children_by_name:
                return current._children_by_name[name]
            if recurseInAnon:
                for child in current._anon_children:
                    if name in child._children_by_name:
                        return child._children_by_name[name]
            if not searchInSiblings:
                break
            current = current.siblingAbove
        return None

    def direct_lookup(self, key: LookupKey) -> Symbol | None:
        if Symbol.debug_lookup:
            Symbol.debug_indent += 1
            Symbol.debug_print('direct_lookup:')
            Symbol.debug_indent += 1
        s = self
        for ident, id_ in key.data:
            s = s._children_by_name.get(ident.name)
            if Symbol.debug_lookup:
                Symbol.debug_print('name:          ', ident.name)
                Symbol.debug_print('id:            ', id_)
                if s is not None:
                    logger.debug(
                        s.to_string(Symbol.debug_indent + 1, addEndNewline=False)
                    )
                else:
                    Symbol.debug_print('not found')
            if s is None:
                break
        if Symbol.debug_lookup:
            Symbol.debug_indent -= 2
        return s

    def find_declaration(
        self, nestedName: ASTNestedName, typ: str, matchSelf: bool, recurseInAnon: bool
    ) -> Symbol | None:
        # templateShorthand: missing template parameter lists for templates is ok
        if Symbol.debug_lookup:
            Symbol.debug_indent += 1
            Symbol.debug_print('find_declaration:')

        def on_missing_qualified_symbol(
            parent_symbol: Symbol, ident: ASTIdentifier
        ) -> Symbol | None:
            return None

        lookup_result = self._symbol_lookup(
            nestedName,
            on_missing_qualified_symbol,
            ancestor_lookup_type=typ,
            match_self=matchSelf,
            recurse_in_anon=recurseInAnon,
            search_in_siblings=False,
        )
        if Symbol.debug_lookup:
            Symbol.debug_indent -= 1
        if lookup_result is None:
            return None

        symbols = list(lookup_result.symbols)
        if len(symbols) == 0:
            return None
        return symbols[0]

    def to_string(self, indent: int, *, addEndNewline: bool = True) -> str:
        res = [Symbol.debug_indent_string * indent]
        if not self.parent:
            res.append('::')
        else:
            if self.ident:
                res.append(self.ident.name)
            else:
                res.append(str(self.declaration))
            if self.declaration:
                res.append(': ')
                if self.isRedeclaration:
                    res.append('!!duplicate!! ')
                res.append(str(self.declaration))
        if self.docname:
            res.extend((
                '\t(',
                self.docname,
                ')',
            ))
        if addEndNewline:
            res.append('\n')
        return ''.join(res)

    def dump(self, indent: int) -> str:
        return ''.join([
            self.to_string(indent),
            *(c.dump(indent + 1) for c in self._children),
        ])



# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/finnish/stop.txt

from __future__ import annotations

FINNISH_STOPWORDS = frozenset({
    'ei',
    'eivät',
    'emme',
    'en',
    'et',
    'ette',
    'että',
    'he',
    'heidän',
    'heidät',
    'heihin',
    'heille',
    'heillä',
    'heiltä',
    'heissä',
    'heistä',
    'heitä',
    'hän',
    'häneen',
    'hänelle',
    'hänellä',
    'häneltä',
    'hänen',
    'hänessä',
    'hänestä',
    'hänet',
    'häntä',
    'itse',
    'ja',
    'johon',
    'joiden',
    'joihin',
    'joiksi',
    'joilla',
    'joille',
    'joilta',
    'joina',
    'joissa',
    'joista',
    'joita',
    'joka',
    'joksi',
    'jolla',
    'jolle',
    'jolta',
    'jona',
    'jonka',
    'jos',
    'jossa',
    'josta',
    'jota',
    'jotka',
    'kanssa',
    'keiden',
    'keihin',
    'keiksi',
    'keille',
    'keillä',
    'keiltä',
    'keinä',
    'keissä',
    'keistä',
    'keitä',
    'keneen',
    'keneksi',
    'kenelle',
    'kenellä',
    'keneltä',
    'kenen',
    'kenenä',
    'kenessä',
    'kenestä',
    'kenet',
    'ketkä',
    'ketä',
    'koska',
    'kuin',
    'kuka',
    'kun',
    'me',
    'meidän',
    'meidät',
    'meihin',
    'meille',
    'meillä',
    'meiltä',
    'meissä',
    'meistä',
    'meitä',
    'mihin',
    'miksi',
    'mikä',
    'mille',
    'millä',
    'miltä',
    'minkä',
    'minua',
    'minulla',
    'minulle',
    'minulta',
    'minun',
    'minussa',
    'minusta',
    'minut',
    'minuun',
    'minä',
    'missä',
    'mistä',
    'mitkä',
    'mitä',
    'mukaan',
    'mutta',
    'ne',
    'niiden',
    'niihin',
    'niiksi',
    'niille',
    'niillä',
    'niiltä',
    'niin',
    'niinä',
    'niissä',
    'niistä',
    'niitä',
    'noiden',
    'noihin',
    'noiksi',
    'noilla',
    'noille',
    'noilta',
    'noin',
    'noina',
    'noissa',
    'noista',
    'noita',
    'nuo',
    'nyt',
    'näiden',
    'näihin',
    'näiksi',
    'näille',
    'näillä',
    'näiltä',
    'näinä',
    'näissä',
    'näistä',
    'näitä',
    'nämä',
    'ole',
    'olemme',
    'olen',
    'olet',
    'olette',
    'oli',
    'olimme',
    'olin',
    'olisi',
    'olisimme',
    'olisin',
    'olisit',
    'olisitte',
    'olisivat',
    'olit',
    'olitte',
    'olivat',
    'olla',
    'olleet',
    'ollut',
    'on',
    'ovat',
    'poikki',
    'se',
    'sekä',
    'sen',
    'siihen',
    'siinä',
    'siitä',
    'siksi',
    'sille',
    'sillä',
    'siltä',
    'sinua',
    'sinulla',
    'sinulle',
    'sinulta',
    'sinun',
    'sinussa',
    'sinusta',
    'sinut',
    'sinuun',
    'sinä',
    'sitä',
    'tai',
    'te',
    'teidän',
    'teidät',
    'teihin',
    'teille',
    'teillä',
    'teiltä',
    'teissä',
    'teistä',
    'teitä',
    'tuo',
    'tuohon',
    'tuoksi',
    'tuolla',
    'tuolle',
    'tuolta',
    'tuon',
    'tuona',
    'tuossa',
    'tuosta',
    'tuota',
    'tähän',
    'täksi',
    'tälle',
    'tällä',
    'tältä',
    'tämä',
    'tämän',
    'tänä',
    'tässä',
    'tästä',
    'tätä',
    'vaan',
    'vai',
    'vaikka',
    'yli',
})



from __future__ import annotations

from importlib import import_module
from typing import TYPE_CHECKING

from sphinx.errors import ExtensionError

if TYPE_CHECKING:
    from typing import Any


def import_object(object_name: str, /, source: str = '') -> Any:
    """Import python object by qualname."""
    obj_path = object_name.split('.')
    module_name = obj_path.pop(0)
    try:
        obj = import_module(module_name)
        for name in obj_path:
            module_name += '.' + name
            try:
                obj = getattr(obj, name)
            except AttributeError:
                obj = import_module(module_name)
    except (AttributeError, ImportError) as exc:
        if source:
            msg = f'Could not import {object_name} (needed for {source})'
            raise ExtensionError(msg, exc) from exc
        msg = f'Could not import {object_name}'
        raise ExtensionError(msg, exc) from exc
    return obj



from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING, cast

from docutils import nodes
from docutils.nodes import make_id
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives import images, tables
from docutils.parsers.rst.directives.misc import Meta

from sphinx.directives import optional_int
from sphinx.locale import __
from sphinx.util import logging
from sphinx.util.docutils import SphinxDirective, _normalize_options
from sphinx.util.nodes import set_source_info
from sphinx.util.osutil import SEP, relpath

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import ClassVar

    from docutils.nodes import Node

    from sphinx.application import Sphinx
    from sphinx.util.typing import ExtensionMetadata, OptionSpec


logger = logging.getLogger(__name__)


class Figure(images.Figure):
    """The figure directive which applies `:name:` option to the figure node
    instead of the image node.
    """

    def run(self) -> Sequence[Node]:
        name = self.options.pop('name', None)
        result = super().run()
        if len(result) == 2 or isinstance(result[0], nodes.system_message):
            return result

        assert len(result) == 1
        figure_node = cast('nodes.figure', result[0])
        if name:
            # set ``name`` to figure_node if given
            self.options['name'] = name
            self.add_name(figure_node)

        # copy lineno from image node
        if figure_node.line is None and len(figure_node) == 2:
            caption = cast('nodes.caption', figure_node[1])
            figure_node.line = caption.line

        return [figure_node]


class CSVTable(tables.CSVTable):
    """The csv-table directive which searches a CSV file from Sphinx project's source
    directory when an absolute path is given via :file: option.
    """

    def run(self) -> Sequence[nodes.table | nodes.system_message]:
        if 'file' in self.options and self.options['file'].startswith((SEP, os.sep)):
            env = self.state.document.settings.env
            filename = Path(self.options['file'])
            if filename.exists():
                logger.warning(
                    __(
                        '":file:" option for csv-table directive now recognizes '
                        'an absolute path as a relative path from source directory. '
                        'Please update your document.'
                    ),
                    location=(env.current_document.docname, self.lineno),
                )
            else:
                abspath = env.srcdir / self.options['file'][1:]
                doc_dir = env.doc2path(env.current_document.docname).parent
                self.options['file'] = relpath(abspath, doc_dir)

        return super().run()


class Code(SphinxDirective):
    """Parse and mark up content of a code block.

    This is compatible with docutils' :rst:dir:`code` directive.
    """

    optional_arguments = 1
    option_spec: ClassVar[OptionSpec] = {
        'class': directives.class_option,
        'force': directives.flag,
        'name': directives.unchanged,
        'number-lines': optional_int,
    }
    has_content = True

    def run(self) -> list[Node]:
        self.assert_has_content()

        self.options = _normalize_options(self.options)
        code = '\n'.join(self.content)
        node = nodes.literal_block(
            code,
            code,
            classes=self.options.get('classes', []),
            force='force' in self.options,
            highlight_args={},
        )
        self.add_name(node)
        set_source_info(self, node)

        if self.arguments:
            # highlight language specified
            node['language'] = self.arguments[0]
        else:
            # no highlight language specified.  Then this directive refers the current
            # highlight setting via ``highlight`` directive or ``highlight_language``
            # configuration.
            node['language'] = (
                self.env.current_document.highlight_language
                or self.config.highlight_language
            )

        if 'number-lines' in self.options:
            node['linenos'] = True

            # if number given, treat as lineno-start.
            if self.options['number-lines']:
                node['highlight_args']['linenostart'] = self.options['number-lines']

        return [node]


class MathDirective(SphinxDirective):
    has_content = True
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec: ClassVar[OptionSpec] = {
        'label': directives.unchanged,
        'name': directives.unchanged,
        'class': directives.class_option,
        'no-wrap': directives.flag,
        'nowrap': directives.flag,
    }

    def run(self) -> list[Node]:
        # Copy the old option name to the new one
        # xref RemovedInSphinx90Warning
        # deprecate nowrap in Sphinx 9.0
        if 'no-wrap' not in self.options and 'nowrap' in self.options:
            self.options['no-wrap'] = self.options['nowrap']

        latex = '\n'.join(self.content)
        if self.arguments and self.arguments[0]:
            latex = self.arguments[0] + '\n\n' + latex
        label = self.options.get('label', self.options.get('name'))
        node = nodes.math_block(
            latex,
            latex,
            classes=self.options.get('class', []),
            docname=self.env.current_document.docname,
            number=None,
            label=label,
        )
        node['no-wrap'] = node['nowrap'] = 'no-wrap' in self.options
        self.add_name(node)
        self.set_source_info(node)

        ret: list[Node] = [node]
        self.add_target(ret)
        return ret

    def add_target(self, ret: list[Node]) -> None:
        node = cast('nodes.math_block', ret[0])

        # assign label automatically if math_number_all enabled
        if node['label'] == '' or (self.config.math_number_all and not node['label']):  # NoQA: PLC1901
            seq = self.env.new_serialno('sphinx.ext.math#equations')
            node['label'] = f'{self.env.current_document.docname}:{seq}'

        # no targets and numbers are needed
        if not node['label']:
            return

        # register label to domain
        domain = self.env.domains.math_domain
        domain.note_equation(
            self.env.current_document.docname, node['label'], location=node
        )
        node['number'] = domain.get_equation_number_for(node['label'])

        # add target node
        node_id = make_id('equation-%s' % node['label'])
        target = nodes.target('', '', ids=[node_id])
        self.state.document.note_explicit_target(target)
        ret.insert(0, target)


class Rubric(SphinxDirective):
    """A patch of the docutils' :rst:dir:`rubric` directive,
    which adds a level option to specify the heading level of the rubric.
    """

    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {
        'class': directives.class_option,
        'name': directives.unchanged,
        'heading-level': lambda c: directives.choice(c, ('1', '2', '3', '4', '5', '6')),
    }

    def run(self) -> list[nodes.rubric | nodes.system_message]:
        self.options = _normalize_options(self.options)
        rubric_text = self.arguments[0]
        textnodes, messages = self.parse_inline(rubric_text, lineno=self.lineno)
        if 'heading-level' in self.options:
            self.options['heading-level'] = int(self.options['heading-level'])
        rubric = nodes.rubric(rubric_text, '', *textnodes, **self.options)
        self.add_name(rubric)
        return [rubric, *messages]


def setup(app: Sphinx) -> ExtensionMetadata:
    directives.register_directive('figure', Figure)
    directives.register_directive('meta', Meta)
    directives.register_directive('csv-table', CSVTable)
    directives.register_directive('code', Code)
    directives.register_directive('math', MathDirective)
    directives.register_directive('rubric', Rubric)

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



# automatically generated by utils/generate-snowball.py
# from https://github.com/snowballstem/snowball-website/raw/efb4ae4d65769fb1652acbe608c0c817e746c730/algorithms/hungarian/stop.txt

from __future__ import annotations

HUNGARIAN_STOPWORDS = frozenset({
    'a',
    'abban',
    'ahhoz',
    'ahogy',
    'ahol',
    'aki',
    'akik',
    'akkor',
    'alatt',
    'amely',
    'amelyek',
    'amelyekben',
    'amelyeket',
    'amelyet',
    'amelynek',
    'ami',
    'amikor',
    'amit',
    'amolyan',
    'amíg',
    'annak',
    'arra',
    'arról',
    'az',
    'azok',
    'azon',
    'azonban',
    'azt',
    'aztán',
    'azután',
    'azzal',
    'azért',
    'be',
    'belül',
    'benne',
    'bár',
    'cikk',
    'cikkek',
    'cikkeket',
    'csak',
    'de',
    'e',
    'ebben',
    'eddig',
    'egy',
    'egyes',
    'egyetlen',
    'egyik',
    'egyre',
    'egyéb',
    'egész',
    'ehhez',
    'ekkor',
    'el',
    'ellen',
    'elsõ',
    'elég',
    'elõ',
    'elõször',
    'elõtt',
    'emilyen',
    'ennek',
    'erre',
    'ez',
    'ezek',
    'ezen',
    'ezt',
    'ezzel',
    'ezért',
    'fel',
    'felé',
    'hanem',
    'hiszen',
    'hogy',
    'hogyan',
    'igen',
    'ill',
    'ill.',
    'illetve',
    'ilyen',
    'ilyenkor',
    'ismét',
    'ison',
    'itt',
    'jobban',
    'jó',
    'jól',
    'kell',
    'kellett',
    'keressünk',
    'keresztül',
    'ki',
    'kívül',
    'között',
    'közül',
    'legalább',
    'legyen',
    'lehet',
    'lehetett',
    'lenne',
    'lenni',
    'lesz',
    'lett',
    'maga',
    'magát',
    'majd',
    'meg',
    'mellett',
    'mely',
    'melyek',
    'mert',
    'mi',
    'mikor',
    'milyen',
    'minden',
    'mindenki',
    'mindent',
    'mindig',
    'mint',
    'mintha',
    'mit',
    'mivel',
    'miért',
    'most',
    'már',
    'más',
    'másik',
    'még',
    'míg',
    'nagy',
    'nagyobb',
    'nagyon',
    'ne',
    'nekem',
    'neki',
    'nem',
    'nincs',
    'néha',
    'néhány',
    'nélkül',
    'olyan',
    'ott',
    'pedig',
    'persze',
    'rá',
    's',
    'saját',
    'sem',
    'semmi',
    'sok',
    'sokat',
    'sokkal',
    'szemben',
    'szerint',
    'szinte',
    'számára',
    'talán',
    'tehát',
    'teljes',
    'tovább',
    'továbbá',
    'több',
    'ugyanis',
    'utolsó',
    'után',
    'utána',
    'vagy',
    'vagyis',
    'vagyok',
    'valaki',
    'valami',
    'valamint',
    'való',
    'van',
    'vannak',
    'vele',
    'vissza',
    'viszont',
    'volna',
    'volt',
    'voltak',
    'voltam',
    'voltunk',
    'által',
    'általában',
    'át',
    'én',
    'éppen',
    'és',
    'így',
    'õ',
    'õk',
    'õket',
    'össze',
    'úgy',
    'új',
    'újabb',
    'újra',
})
